# Python module

This version of `pygio` is pip-installable and works with `mpi4py`.

## Requirements

Currently, a CMake version >= 3.11.0 is required to fetch dependencies during configuration. 
The `pygio` module also requires MPI libraries to be findable by CMake's FindMPI. The compiler
needs to support C++17 (make sure that `CC` and `CXX` point to the correct compiler)

## Install

The python library can be installed by running pip in the **main folder**:
```bash
pip install .
```
It will use the compiler referred by the `CC` and `CXX` environment variable. If the compiler
supports OpenMP, the library will be threaded. Make sure to set `OMP_NUM_THREADS` to an
appropriate variable, in particluar when using multiple MPI ranks per node.

## Usage

The library can then be imported in python. Here is a small example script:
```python
import numpy as np
import pygio

# inspect file
pygio.inspect_genericio("generic_io_file")

# read all variables
data = pygio.read_genericio("generic_io_file")

# read only a subset of variables
data_partial = pygio.read_genericio("generic_io_file", ["x", "y", "z"])
data_x = data_partial["x"]

# write data to file
pygio.write_genericio("new_generic_io_file", 
     variables = {"x": np.ones(100), "y": np.ones(100)},
     phys_scale = [1, 1, 1],
     phys_origin = [0, 0, 0],
     method = PyGenericIO.FileIO.FileIOPOSIX
     )


### USING THE CLASS BASED INTERFACE ###

# instantiate a GenericIO class
gio_file = pygio.PyGenericIO("generic_io_file")

# inspect (prints to python stdout, also works in notebook)
gio_file.inspect()

# get variables
gio_vars = gio_file.get_variables()

# print variable names
for var in gio_vars:
    print(var.name, var.size, var.element_size, var.is_float)
    
# read data
data_all = gio_file.read()
data_partial = gio_file.read(["x", "y", "z"])
```

Further methods and members of GenericIO can easly be interfaced by editing `python_new/genericio.cpp`.


### Using MPI
The `pygio` library is working with mpi4py. Here is an example file:
```python
from mpi4py import MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
ranks = comm.Get_size()

import numpy as np
import pygio

# read locally
data = pygio.read_genericio("generic_io_file")
# get local number of elements from the first element in dictionary
num_elems = len(next(iter(data.values())))
# reduce total number of elements
num_elems_total = comm.allreduce(num_elems)
if rank == 0:
    print(f"Reading file with {ranks} ranks")
    print(f"Total number of particles: {num_elems_total}")
    print("The data contains the following variables:")
    for k, d in data.items():
        print(f"\t{k:5s}, dtype={d.dtype}")

for i in range(ranks):
    if i == rank:
        print(f"rank {rank} read {num_elems} elements")
    comm.Barrier()
```

It can be executed with `mpirun`:
```bash
mpirun -n 8 python testpygio_mpi.py
```

Here is an output for 1 and 8 ranks:
```
Reading file with 1 ranks
Total number of particles: 562500
The data contains the following variables:
        id   , dtype=int64
        vx   , dtype=float32
        vy   , dtype=float32
        vz   , dtype=float32
        x    , dtype=float32
        y    , dtype=float32
        z    , dtype=float32
rank 0 read 562500 elements
```

```
Reading file with 8 ranks
Total number of particles: 562500
The data contains the following variables:
        id   , dtype=int64
        vx   , dtype=float32
        vy   , dtype=float32
        vz   , dtype=float32
        x    , dtype=float32
        y    , dtype=float32
        z    , dtype=float32
rank 0 read 70000 elements
rank 1 read 70000 elements
rank 2 read 70000 elements
rank 3 read 70000 elements
rank 4 read 70625 elements
rank 5 read 70625 elements
rank 6 read 70625 elements
rank 7 read 70625 elements
```

### Force the non-MPI version to be imported

Some clusters prevent the loading of MPI libraries on the login nodes. In order
to still be able to use the same `pygio` on the login nodes, the non-MPI library
can be loaded by setting the `GENERICIO_NO_MPI` environment variable before
importing `pygio`:

```python
import os
os.environ['GENERICIO_NO_MPI'] = 'True'
import pygio
```

A warning will be printed that the writing capabilities of genericio are not
available in this mode.